==> tclMotif.1.3.tar.gz: tm.1.3/src/tmUtils.c <==
/*
 * tmUtils.c --
 *	This module contains general purpose routines
 *	used by the Tm toolkit.
 *
 * Copyright 1993 Jan Newmarch, University of Canberra.
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The author
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

#include <sys/time.h>
#include <ctype.h>
#include <X11/keysym.h>
#include "tm.h"
#include "tmFuncs.h"

/*
 *--------------------------------------------------------------
 *
 * Tm_StoreWidgetInfo --
 *
 *	create a hash table entry for a new widget, with useful
 *	info in it.
 *
 * Results:
 *
 *	modifies hash table in "interp"
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */

void
Tm_StoreWidgetInfo(path, w, interp)
    char *path;
    Tm_Widget *w;
    Tcl_Interp *interp;
{
    int new;
    Tcl_HashEntry *hPtr;

/*
    hPtr = Tcl_CreateHashEntry(&WidgetTable, path, &new);
    Tcl_SetHashValue(hPtr, (ClientData) w);
*/
}


/*
 *--------------------------------------------------------------
 *
 * Tm_NameFromPath --
 *
 *	find the part of the path after the last '.'
 *
 * Results:
 *
 *	returns a pointer to the name
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */

char *
Tm_NameFromPath(pathName)
    char *pathName;
{
    char *p;

    if ((p = strrchr(pathName, '.')) == NULL) {
	return pathName;
    } else {
	return (p + 1);
    }
}



/*
 *--------------------------------------------------------------
 *
 * Tm_HiddenParentPath --
 *
 *	Create a path based on current path that cannot be duplicated
 *	actually, makes "first.last" into "first..last"
 *
 * Results:
 *
 *	returns a pointer to the name
 *
 * Side effects:
 *
 *	allocates memory
 *--------------------------------------------------------------
 */

char *
Tm_HiddenParentPath(path)
    char *path;
{
    char *p;
    char *parent;
    int len;

    p = strrchr(path, '.');
    if (p == NULL) {
        return NULL;
    }
    len = p - path + 1;

    parent = XtMalloc(strlen(path) + 2);
    strncpy(parent, path, len);
    parent[len] = '.';
    strcpy(parent + len + 1, p + 1);

    return parent;
}


/*
 *--------------------------------------------------------------
 *
 * Tm_HiddenParentPath --
 *
 *	Create a path based on current path that cannot be duplicated
 *	actually, makes "first.last" into "first..last"
 *
 * Results:
 *
 *	returns a pointer to the name
 *
 * Side effects:
 *
 *	allocates memory
 *--------------------------------------------------------------
 */

char *
Tm_ParentPath(path)
    char *path;
{
    char *p;
    char *parent;
    int len;

    p = strrchr(path, '.');
    if (p == NULL) {
        return NULL;
    }
    if (p == path)
	len = 1; /* parent is `.' */
    else
        len = p - path;

    parent = XtMalloc(len + 1);
    strncpy(parent, path, len);
    parent[len] = '\0';

    return parent;
}

#if 0
/*
 *--------------------------------------------------------------
 *
 * Tm_ExtractArg --
 *	find a name and its value in a list
 *
 * Results:
 *	returns a pointer to the value
 *
 * Side effects:
 *	none
 *--------------------------------------------------------------
 */

static char *
Tm_ExtractArg(name, argc, argv)
    char *name;
    int argc;
    char *argv[];
{
   int n;

    for (n = 0; n < argc; n++) {
	if (strcmp(name, argv[n]) == 0) {
	    if (++n < argc) {
		return argv[n];
	    } else {
		return NULL;
	    }
	}
    }
    return NULL;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_MakeXEvent --
 *
 *	Create an X event that can be used to send to another window
 *
 * Results:
 *	a Tcl result value
 *
 * Side effects:
 *	returns a filled in event in xev
 *
 * Restrictions:
 *	only handles a small number of X types
 *--------------------------------------------------------------
 */
int
Tm_MakeXEvent(w, interp, xev, argc, argv)
    Widget w;
    Tcl_Interp *interp;
    XEvent *xev;
    int argc;
    char **argv;
{
    char *type;
    char *value;
    struct timeval  tp;
    Display *display;


    /* let them know we made this thing up */
    xev->xany.send_event = True;

    /* fill in the rest of the xany values.
     */
    xev->xany.display = display = XtDisplay(w);
    xev->xany.window = XtWindow(w);

    if ((type = Tm_ExtractArg("-type", argc, argv)) == NULL) {
        type = "ClientMessage";
    }

    if (type[0] == 'B') {
        if (strcmp(type, "ButtonRelease") == 0) {
	    xev->type = ButtonRelease;
	} else
	if (strcmp(type, "ButtonPress") == 0) {
	    xev->type = ButtonPress;
	} else {
	    sprintf(interp->result, "illegal type %s\n", type);
	    return TCL_ERROR;
	}

	xev->xbutton.time = XtLastTimestampProcessed(display);

	if ((value = Tm_ExtractArg("-x", argc, argv)) != NULL) {
	    xev->xbutton.x = atoi(value);
	} else {
	    xev->xbutton.x = 0;
	}

	if ((value = Tm_ExtractArg("-y", argc, argv)) != NULL) {
	    xev->xbutton.y = atoi(value);
	} else {
	    xev->xbutton.y = 0;
	}

	return TCL_OK;
    }

    if (type[0] == 'K') {
	unsigned int mask;
	KeySym keysym;
	KeyCode keycode;

        if (strcmp(type, "KeyRelease") == 0) {
	    xev->type = KeyRelease;
	} else
	if (strcmp(type, "KeyPress") == 0) {
	    xev->type = KeyPress;
	} else {
	    sprintf(interp->result, "illegal type %s\n", type);
	    return TCL_ERROR;;
	}

	xev->xkey.time = XtLastTimestampProcessed(display);

	if ((value = Tm_ExtractArg("-x", argc, argv)) != NULL) {
	    xev->xkey.x = atoi(value);
	} else {
	    xev->xkey.x = 0;
	}

	if ((value = Tm_ExtractArg("-y", argc, argv)) != NULL) {
	    xev->xkey.y = atoi(value);
	} else {
	    xev->xkey.y = 0;
	}

	if ((value = Tm_ExtractArg("-keysym", argc, argv)) != NULL) {
	    if ((keysym = XStringToKeysym(value)) == NoSymbol) {
		sprintf(interp->result, "illegal keysym\"%s\"", value);
		return TCL_ERROR;
	    }
	    if ((keycode = XKeysymToKeycode(display, keysym)) == 0) {
		sprintf(interp->result, "no keycode for keysym \"%s\"", value);
		return TCL_ERROR;
	    }

	    mask = 0;
	    /* is this what we should do? */
	    if (keysym >= XK_A && keysym <= XK_Z)
		mask |= ShiftMask;
	    
	    xev->xkey.keycode = keycode;
	    xev->xkey.state = mask;
	} else {
	    sprintf(interp->result, "Key event must have a keysym");
	    return TCL_ERROR;
	}

	return TCL_OK;
    }

    if (strcmp(type, "ClientMessage") == 0) {
	return TCL_OK;
    }

    sprintf(interp->result, "illegal type %s\n", type);
    return TCL_ERROR;;
}
#endif

/*
 *--------------------------------------------------------------
 *
 * Tm_EndToken --
 *
 *	return the next char past a string or quoted string
 *	using the actions parameter syntax (Asente & Swick appendix B)
 *
 * Results:
 *	return pointer to next char after string
 *
 *--------------------------------------------------------------
 */

char *
Tm_EndToken(s)
    char *s;
{
    if (s == NULL || *s == '\0')
	return s;

    /* quoted string */
    if (*s == '"') {
	s++;
	while (*s != '"' && *s != '\0')
	    s++;
	if (*s == '"')
	    return (s+1);
	else
	    return s;
    }

    /* unquoted string */
    while (*s != ' ' && *s != '\t' && *s != ',' &&
	   *s != '\n' && *s != ')' && *s != '\0')
	s++;
    return s;
}
/*
 *--------------------------------------------------------------
 *
 * Tm_ParseAction --
 *
 *	Parse an action into action name + args
 *
 * Results:
 *	sets the action, options and num_options
 *
 * Side-effects:
 * 	none
 *
 * Caveats:
 *	Somehow, this is an ugly piece of code. Tidy it up!
 *--------------------------------------------------------------
 */

int
Tm_ParseAction(orig, action, params, num_params)
    char *orig;
    char **action;
    char *params[];
    Cardinal *num_params;
{
    int n = 0;
    char *start, *end;
#   define skipblanks(p) while (isspace(*p)) p++;

    skipblanks(orig);
    /* make a new string, and put nulls in it at the end of tokens */
    start = XtNewString(orig);
    *action = end = start;

    /* delimit action name */
    while (isalnum(*end) || *end == '_' || *end == '-')
	end++;
    if (*end != '(') {
	fprintf(stderr, "no \"(\" after name\n");
	return TCL_ERROR;
    }
    *end = '\0';

    start = end + 1;
    for (n = 0; n < TM_NUM_PARAMS; True) {
	skipblanks(start);
	if (*start == ')')
	    break;

	/* new token */
	end = Tm_EndToken(start);

	/* lose quotes if it was a quoted string */
	if (*start == '"') {
	    start++;
	    *(end - 1) = '\0';
	}
	params[n++] = start;

	if (isspace(*end)) {
	    *end++ = '\0';
	}
	skipblanks(end);
	if (*end == '\0') {
	    fprintf(stderr, "unterminated action\n");
	    return TCL_ERROR;
	}
	if (*end == ')') {
	    *end = '\0';
	    break;
	}
	if (*end != ',') {
	    fprintf(stderr, "no \",\" delimiter\n");
	    return TCL_ERROR;
	}
	*end = '\0';
	start = end + 1;
    }

    *num_params = n;
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_ClassToTmCmd --
 *
 *	map from a widget's class to its command info struct
 *
 *--------------------------------------------------------------
 */
Tm_Cmd *
Tm_ClassToWidgetCommandProc(clas)
    WidgetClass clas;
{
    register Tm_Cmd *cmdPtr;
    char *command = NULL;
    int n;

    /* first find the command name from the command <-> class arrays */

    /* check if it is one of the tclMotif commands */
    n = 0;
    while (Tm_CommToClass[n].command != NULL) {
        if ( *(Tm_CommToClass[n].clas) ==  clas) {
            command = Tm_CommToClass[n].command;
            break;
        }
        n++;
    }

    /* check if it is one of the external widgets */
    if (command == NULL) {
        n = 0;
        while (Tm_ExternCommandToClass[n].command != NULL) {
            if ( *(Tm_ExternCommandToClass[n].clas) == clas) {
                command = Tm_ExternCommandToClass[n].command;
                break;
            }
            n++;
        }
    }
    if (command == NULL) {
	return NULL;
    }

    /* now find the command info from the Tm_Cmd arrays */
    for (cmdPtr = Tm_Commands; cmdPtr->name != NULL; cmdPtr++) {
	if (strcmp(command, cmdPtr->name) == 0) {
	    return cmdPtr;
	}
    }

    for (cmdPtr = Tm_ExternCommands; cmdPtr->name != NULL; cmdPtr++) {
	if (strcmp(command, cmdPtr->name) == 0) {
	    return cmdPtr;
	}
    }

    return NULL;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_CreateTmInfo --
 *
 *	create a Tm_Widget record and place it in the widget
 *
 *--------------------------------------------------------------
 */

int
Tm_CreateTmInfo(interp, w, path, parent, displayInfo)
    Tcl_Interp *interp;
    Widget w;
    char *path;
    char *parent;
    Tm_Display *displayInfo;
{
    Tm_Widget *wPtr;
    WidgetClass clas;
    Tm_Cmd *cmdPtr;

    wPtr = (Tm_Widget *) XtMalloc (sizeof (Tm_Widget));
    wPtr -> interp = interp;
    wPtr -> widget = w;
    wPtr -> pathName = XtNewString(path);
    wPtr -> parent = parent;
    wPtr -> dropProc = NULL;
    wPtr -> displayInfo = displayInfo;

    Tm_StoreWidgetInfo(path, wPtr, interp);
    XtVaSetValues(w, XmNuserData, wPtr, NULL);

    /* create a different command for each widget type
     */

    cmdPtr = Tm_ClassToWidgetCommandProc(XtClass(w));
    if (cmdPtr != NULL) {
	Tcl_CreateCommand (interp, path, cmdPtr->widgetCmdProc,
                 (ClientData) wPtr, (void (*) ()) NULL);
    } else {
        Tcl_CreateCommand (interp, path, Tm_AnyWidgetCmd,
                 (ClientData) wPtr, (void (*) ()) NULL);
    }

    XtAddCallback(w, XmNdestroyCallback, Tm_DestroyWidgetHandler,
        (XtPointer) wPtr);

   Tcl_SetResult(interp, wPtr -> pathName, TCL_VOLATILE);

    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * Tm_CreateTmInfoInChildren --
 *
 *      create a Tm_Widget record in each child of the widget
 *
 *--------------------------------------------------------------
 */

void 
Tm_CreateTmInfoInChildren(interp, w, path, displayInfo)
    Tcl_Interp *interp;
    Widget w;
    char *path;
    Tm_Display *displayInfo;
{
    Widget *children;
    Cardinal num_children;
    Cardinal n;
    char *name;
    Widget child;
    char *child_path;

    XtVaGetValues(w, XmNchildren, &children, 
		     XmNnumChildren, &num_children,
		     NULL);

    for (n = 0; n < num_children; n++) {
	child = children[n];
	name = XtName(child);
	if (strlen(name) == 0) {
	    /* are there any of these? be safe */
	    continue;
	}
	child_path = XtMalloc(strlen(path) + strlen(name) + 2);
	strcpy(child_path, path);
	strcat(child_path, ".");
	strcat(child_path, name);
	Tm_CreateTmInfo(interp, child, child_path, path, displayInfo);
	
	/* and recurse down tree */
	if (XtIsComposite(child)) {
	    Tm_CreateTmInfoInChildren(interp, child, child_path,
			displayInfo);
	}
	XtFree(child_path);
    }
}

==> tclMotif.1.4.tar.gz: tm.1.4/src/tmUtils.c <==
/*
 * tmUtils.c --
 *	This module contains general purpose routines
 *	used by the Tm toolkit.
 *
 * Copyright 1993 Jan Newmarch, University of Canberra.
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The author
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

#include <sys/time.h>
#include <ctype.h>
#include <X11/keysym.h>
#include <Xm/CascadeB.h>
#include <Xm/CascadeBG.h>
#include "tm.h"
#include "tmFuncs.h"

/*
 *--------------------------------------------------------------
 *
 * Tm_StoreWidgetInfo --
 *
 *	create a hash table entry for a new widget, with useful
 *	info in it.
 *
 * Results:
 *
 *	modifies hash table in "interp"
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */

void
Tm_StoreWidgetInfo(path, w, interp)
    char *path;
    Tm_Widget *w;
    Tcl_Interp *interp;
{
/*
    int new;
    Tcl_HashEntry *hPtr;

    hPtr = Tcl_CreateHashEntry(&WidgetTable, path, &new);
    Tcl_SetHashValue(hPtr, (ClientData) w);
*/
}


/*
 *--------------------------------------------------------------
 *
 * Tm_NameFromPath --
 *
 *	find the part of the path after the last '.'
 *
 * Results:
 *
 *	returns a pointer to the name
 *
 * Side effects:
 *
 *--------------------------------------------------------------
 */

char *
Tm_NameFromPath(pathName)
    char *pathName;
{
    char *p;

    if ((p = strrchr(pathName, '.')) == NULL) {
	return pathName;
    } else {
	return (p + 1);
    }
}



/*
 *--------------------------------------------------------------
 *
 * Tm_HiddenParentPath --
 *
 *	Create a path based on current path that cannot be duplicated
 *	actually, makes "first.last" into "first..last"
 *
 * Results:
 *
 *	returns a pointer to the name
 *
 * Side effects:
 *
 *	allocates memory
 *--------------------------------------------------------------
 */

char *
Tm_HiddenParentPath(path)
    char *path;
{
    char *p;
    char *parent;
    int len;

    p = strrchr(path, '.');
    if (p == NULL) {
        return NULL;
    }
    len = p - path + 1;

    parent = XtMalloc(strlen(path) + 2);
    strncpy(parent, path, len);
    parent[len] = '.';
    strcpy(parent + len + 1, p + 1);

    return parent;
}


/*
 *--------------------------------------------------------------
 *
 * Tm_HiddenParentPath --
 *
 *	Create a path based on current path that cannot be duplicated
 *	actually, makes "first.last" into "first..last"
 *
 * Results:
 *
 *	returns a pointer to the name
 *
 * Side effects:
 *
 *	allocates memory
 *--------------------------------------------------------------
 */

char *
Tm_ParentPath(path)
    char *path;
{
    char *p;
    char *parent;
    int len;

    p = strrchr(path, '.');
    if (p == NULL) {
        return NULL;
    }
    if (p == path)
	len = 1; /* parent is `.' */
    else
        len = p - path;

    parent = XtMalloc(len + 1);
    strncpy(parent, path, len);
    parent[len] = '\0';

    return parent;
}

#if 0
/*
 *--------------------------------------------------------------
 *
 * Tm_ExtractArg --
 *	find a name and its value in a list
 *
 * Results:
 *	returns a pointer to the value
 *
 * Side effects:
 *	none
 *--------------------------------------------------------------
 */

static char *
Tm_ExtractArg(name, argc, argv)
    char *name;
    int argc;
    char *argv[];
{
   int n;

    for (n = 0; n < argc; n++) {
	if (strcmp(name, argv[n]) == 0) {
	    if (++n < argc) {
		return argv[n];
	    } else {
		return NULL;
	    }
	}
    }
    return NULL;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_MakeXEvent --
 *
 *	Create an X event that can be used to send to another window
 *
 * Results:
 *	a Tcl result value
 *
 * Side effects:
 *	returns a filled in event in xev
 *
 * Restrictions:
 *	only handles a small number of X types
 *--------------------------------------------------------------
 */
int
Tm_MakeXEvent(w, interp, xev, argc, argv)
    Widget w;
    Tcl_Interp *interp;
    XEvent *xev;
    int argc;
    char **argv;
{
    char *type;
    char *value;
    struct timeval  tp;
    Display *display;


    /* let them know we made this thing up */
    xev->xany.send_event = True;

    /* fill in the rest of the xany values.
     */
    xev->xany.display = display = XtDisplay(w);
    xev->xany.window = XtWindow(w);

    if ((type = Tm_ExtractArg("-type", argc, argv)) == NULL) {
        type = "ClientMessage";
    }

    if (type[0] == 'B') {
        if (strcmp(type, "ButtonRelease") == 0) {
	    xev->type = ButtonRelease;
	} else
	if (strcmp(type, "ButtonPress") == 0) {
	    xev->type = ButtonPress;
	} else {
	    sprintf(interp->result, "illegal type %s\n", type);
	    return TCL_ERROR;
	}

	xev->xbutton.time = XtLastTimestampProcessed(display);

	if ((value = Tm_ExtractArg("-x", argc, argv)) != NULL) {
	    xev->xbutton.x = atoi(value);
	} else {
	    xev->xbutton.x = 0;
	}

	if ((value = Tm_ExtractArg("-y", argc, argv)) != NULL) {
	    xev->xbutton.y = atoi(value);
	} else {
	    xev->xbutton.y = 0;
	}

	return TCL_OK;
    }

    if (type[0] == 'K') {
	unsigned int mask;
	KeySym keysym;
	KeyCode keycode;

        if (strcmp(type, "KeyRelease") == 0) {
	    xev->type = KeyRelease;
	} else
	if (strcmp(type, "KeyPress") == 0) {
	    xev->type = KeyPress;
	} else {
	    sprintf(interp->result, "illegal type %s\n", type);
	    return TCL_ERROR;;
	}

	xev->xkey.time = XtLastTimestampProcessed(display);

	if ((value = Tm_ExtractArg("-x", argc, argv)) != NULL) {
	    xev->xkey.x = atoi(value);
	} else {
	    xev->xkey.x = 0;
	}

	if ((value = Tm_ExtractArg("-y", argc, argv)) != NULL) {
	    xev->xkey.y = atoi(value);
	} else {
	    xev->xkey.y = 0;
	}

	if ((value = Tm_ExtractArg("-keysym", argc, argv)) != NULL) {
	    if ((keysym = XStringToKeysym(value)) == NoSymbol) {
		sprintf(interp->result, "illegal keysym\"%s\"", value);
		return TCL_ERROR;
	    }
	    if ((keycode = XKeysymToKeycode(display, keysym)) == 0) {
		sprintf(interp->result, "no keycode for keysym \"%s\"", value);
		return TCL_ERROR;
	    }

	    mask = 0;
	    /* is this what we should do? */
	    if (keysym >= XK_A && keysym <= XK_Z)
		mask |= ShiftMask;
	    
	    xev->xkey.keycode = keycode;
	    xev->xkey.state = mask;
	} else {
	    sprintf(interp->result, "Key event must have a keysym");
	    return TCL_ERROR;
	}

	return TCL_OK;
    }

    if (strcmp(type, "ClientMessage") == 0) {
	return TCL_OK;
    }

    sprintf(interp->result, "illegal type %s\n", type);
    return TCL_ERROR;;
}
#endif

/*
 *--------------------------------------------------------------
 *
 * Tm_EndToken --
 *
 *	return the next char past a string or quoted string
 *	using the actions parameter syntax (Asente & Swick appendix B)
 *
 * Results:
 *	return pointer to next char after string
 *
 *--------------------------------------------------------------
 */

char *
Tm_EndToken(s)
    char *s;
{
    if (s == NULL || *s == '\0')
	return s;

    /* quoted string */
    if (*s == '"') {
	s++;
	while (*s != '"' && *s != '\0')
	    s++;
	if (*s == '"')
	    return (s+1);
	else
	    return s;
    }

    /* unquoted string */
    while (*s != ' ' && *s != '\t' && *s != ',' &&
	   *s != '\n' && *s != ')' && *s != '\0')
	s++;
    return s;
}
/*
 *--------------------------------------------------------------
 *
 * Tm_ParseAction --
 *
 *	Parse an action into action name + args
 *
 * Results:
 *	sets the action, options and num_options
 *
 * Side-effects:
 * 	none
 *
 * Caveats:
 *	Somehow, this is an ugly piece of code. Tidy it up!
 *--------------------------------------------------------------
 */

int
Tm_ParseAction(orig, action, params, num_params)
    char *orig;
    char **action;
    char *params[];
    Cardinal *num_params;
{
    int n = 0;
    char *start, *end;
#   define skipblanks(p) while (isspace(*p)) p++;

    skipblanks(orig);
    /* make a new string, and put nulls in it at the end of tokens */
    start = XtNewString(orig);
    *action = end = start;

    /* delimit action name */
    while (isalnum(*end) || *end == '_' || *end == '-')
	end++;
    if (*end != '(') {
	fprintf(stderr, "no \"(\" after name\n");
	return TCL_ERROR;
    }
    *end = '\0';

    start = end + 1;
    for (n = 0; n < TM_NUM_PARAMS;) {
	skipblanks(start);
	if (*start == ')')
	    break;

	/* new token */
	end = Tm_EndToken(start);

	/* lose quotes if it was a quoted string */
	if (*start == '"') {
	    start++;
	    *(end - 1) = '\0';
	}
	params[n++] = start;

	if (isspace(*end)) {
	    *end++ = '\0';
	}
	skipblanks(end);
	if (*end == '\0') {
	    fprintf(stderr, "unterminated action\n");
	    return TCL_ERROR;
	}
	if (*end == ')') {
	    *end = '\0';
	    break;
	}
	if (*end != ',') {
	    fprintf(stderr, "no \",\" delimiter\n");
	    return TCL_ERROR;
	}
	*end = '\0';
	start = end + 1;
    }

    *num_params = n;
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_ClassToTmCmd --
 *
 *	map from a widget's class to its command info struct
 *
 *--------------------------------------------------------------
 */
Tm_Cmd *
Tm_ClassToWidgetCommandProc(clas)
    WidgetClass clas;
{
    register Tm_Cmd *cmdPtr;
    char *command = NULL;
    int n;

    /* first find the command name from the command <-> class arrays */

    /* check if it is one of the tclMotif commands */
    n = 0;
    while (Tm_CommToClass[n].command != NULL) {
        if ( *(Tm_CommToClass[n].clas) ==  clas) {
            command = Tm_CommToClass[n].command;
            break;
        }
        n++;
    }

    /* check if it is one of the external widgets */
    if (command == NULL) {
        n = 0;
        while (Tm_ExternCommandToClass[n].command != NULL) {
            if ( *(Tm_ExternCommandToClass[n].clas) == clas) {
                command = Tm_ExternCommandToClass[n].command;
                break;
            }
            n++;
        }
    }
    if (command == NULL) {
	return NULL;
    }

    /* now find the command info from the Tm_Cmd arrays */
    for (cmdPtr = Tm_Commands; cmdPtr->name != NULL; cmdPtr++) {
	if (strcmp(command, cmdPtr->name) == 0) {
	    return cmdPtr;
	}
    }

    for (cmdPtr = Tm_ExternCommands; cmdPtr->name != NULL; cmdPtr++) {
	if (strcmp(command, cmdPtr->name) == 0) {
	    return cmdPtr;
	}
    }

    return NULL;
}

/*
 *--------------------------------------------------------------
 *
 * Tm_CreateTmInfo --
 *
 *	create a Tm_Widget record and place it in the widget
 *
 *--------------------------------------------------------------
 */

int
Tm_CreateTmInfo(interp, w, path, parent, displayInfo)
    Tcl_Interp *interp;
    Widget w;
    char *path;
    char *parent;
    Tm_Display *displayInfo;
{
    Tm_Widget *wPtr;
    Tm_Cmd *cmdPtr;

    wPtr = (Tm_Widget *) XtMalloc (sizeof (Tm_Widget));
    wPtr -> interp = interp;
    wPtr -> widget = w;
    wPtr -> pathName = XtNewString(path);
    wPtr -> parent = XtNewString(parent);
    wPtr -> dropProc = NULL;
    wPtr -> displayInfo = displayInfo;

    Tm_StoreWidgetInfo(path, wPtr, interp);
    XtVaSetValues(w, XmNuserData, wPtr, NULL);

    /* create a different command for each widget type
     */
    cmdPtr = Tm_ClassToWidgetCommandProc(XtClass(w));
    if (cmdPtr != NULL) {
	Tcl_CreateCommand (interp, path, cmdPtr->widgetCmdProc,
                 (ClientData) wPtr, (void (*) ()) NULL);
    } else {
        Tcl_CreateCommand (interp, path, Tm_AnyWidgetCmd,
                 (ClientData) wPtr, (void (*) ()) NULL);
    }
    XtAddCallback(w, XmNdestroyCallback, Tm_DestroyWidgetHandler,
		  (XtPointer) wPtr);
    Tcl_SetResult(interp, wPtr -> pathName, TCL_VOLATILE);
    return TCL_OK;
}


/*
 *--------------------------------------------------------------
 *
 * Tm_CreateTmInfoInChildren --
 *
 *      create a Tm_Widget record in each child of the widget
 *
 *--------------------------------------------------------------
 */

void 
Tm_CreateTmInfoInChildren(interp, w, path, displayInfo)
    Tcl_Interp *interp;
    Widget w;
    char *path;
    Tm_Display *displayInfo;
{
    Widget *children;
    Cardinal num_children;
    Cardinal n;
    char *name;
    Widget child;
    char *child_path;

    XtVaGetValues(w, XmNchildren, &children, 
		     XmNnumChildren, &num_children,
		     NULL);

    for (n = 0; n < num_children; n++) {
	child = children[n];
	name = XtName(child);
	if (strlen(name) == 0) {
	    /* are there any of these? be safe */
	    continue;
	}
	child_path = XtMalloc(strlen(path) + strlen(name) + 2);
	strcpy(child_path, path);
	strcat(child_path, ".");
	strcat(child_path, name);
	Tm_CreateTmInfo(interp, child, child_path, path, displayInfo);
	
	/* and recurse down tree */
	if (XtIsComposite(child))
	    Tm_CreateTmInfoInChildren(interp, child, child_path, displayInfo);
	else
	    Tm_CreateTmMenuInfo(interp,child,path,displayInfo);
	XtFree(child_path);
    }
}

void
Tm_CreateTmMenuInfo(interp, w, parent, displayInfo)
    Tcl_Interp *interp;
    Widget w;
    char *parent;
    Tm_Display *displayInfo;
{
    if (XmIsCascadeButton(w) || XmIsCascadeButtonGadget(w)) {
	Widget mpane = (Widget)NULL;
	
	XtVaGetValues(w,XmNsubMenuId,&mpane,NULL);
	if (mpane != (Widget)NULL) {
	    char *name = XtName(mpane);
	    char *mpath;
	    if (strlen(name) == 0)
		return;				  /* we skip over these */
	    mpath = XtMalloc(strlen(parent) + strlen(name) + 2);
	    strcpy(mpath,parent);
	    /* special case root widget `.' */
	    if (strcmp(parent, ".") != 0) {
		strcat(mpath, ".");
	    }
	    strcat(mpath, name);
	    Tm_CreateTmInfo(interp,mpane,mpath,parent,displayInfo);
	    /* now recurse down the tree */
	    if (XtIsComposite(mpane))
		Tm_CreateTmInfoInChildren(interp,mpane,mpath,displayInfo);
	    XtFree(mpath);
	}
    }
}
